from urllib.parse import urljoin

from selenium.webdriver.common.by import By


def test_webelement_click_at(selenium, website):
    selenium.get(urljoin(website, "/elements.htm"))
    selenium.find_element(By.ID, "beta").click_at(20, 20)
    assert (
        selenium.find_element(By.ID, "message").text == "beta"
    ), "click on element failed"
    selenium.find_element(By.ID, "lite").click_at(-170, -170)
    assert (
        selenium.find_element(By.ID, "message").text == "firefox"
    ), "click on element failed"


def test_webelement_move_to(selenium, website):
    selenium.get(urljoin(website, "/elements.htm"))
    selenium.find_element(By.ID, "developer").move_to()
    assert (
        selenium.find_element(By.ID, "message").text == "overdeveloper"
    ), "click on element failed"


def test_webelement_move_at(selenium, website):
    selenium.get(urljoin(website, "/elements.htm"))
    selenium.find_element(By.ID, "firefox").move_at(40, 40)
    assert (
        selenium.find_element(By.ID, "message").text == "overfirefox"
    ), "click on element failed"
    selenium.find_element(By.ID, "lite").move_at(100, -300)
    assert (
        selenium.find_element(By.ID, "message").text == "overbeta"
    ), "click on element failed"
    selenium.find_element(By.ID, "developer").move_at(300, 10)
    assert (
        selenium.find_element(By.ID, "message").text == "overlite"
    ), "click on element failed"



from urllib.parse import urljoin

import pytest
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By


def test_find_image_no_element(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    with pytest.raises(NoSuchElementException):
        selenium.find_image("tests/html/browser0.png")


def test_find_image_one_element(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    selenium.find_image("tests/html/browser_edge.png")


def test_find_image_three_elements(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    selenium.find_image("tests/html/browser_firefox.png")


def test_find_image_one_element_click(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    element = selenium.find_image("tests/html/browser_edge.png")
    element.click()
    assert (
        selenium.find_element(By.ID, "message").text == "edge1"
    ), "click on element failed"


def test_find_image_one_element_click_at(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    element = selenium.find_image("tests/html/browser_edge.png")
    element.click_at(-100, 200)
    assert (
        selenium.find_element(By.ID, "message").text == "firefox2"
    ), "click on element failed"


def test_find_image_one_element_move_to(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    element = selenium.find_image("tests/html/browser_edge.png")
    element.move_to()
    assert (
        selenium.find_element(By.ID, "message").text == "overedge1"
    ), "click on element failed"


def test_find_image_one_element_move_at(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    element = selenium.find_image("tests/html/browser_edge.png")
    element.move_at(200, 200)
    assert (
        selenium.find_element(By.ID, "message").text == "overchrome2"
    ), "click on element failed"


def test_find_image_one_element_need_wait(selenium, website):
    selenium.get(urljoin(website, "/browsers_delay.htm"))
    with pytest.raises(NoSuchElementException):
        selenium.find_image("tests/html/browser_edge.png")


def test_find_image_one_element_with_wait(selenium, website):
    selenium.implicitly_wait(6)
    selenium.get(urljoin(website, "/browsers_delay.htm"))
    selenium.find_image("tests/html/browser_edge.png")


def test_find_image_click_need_scroll(selenium, website):
    selenium.get(urljoin(website, "/browsers_scroll.htm"))
    with pytest.raises(NoSuchElementException):
        selenium.find_image("tests/html/browser_edge.png")


def test_find_image_click_with_scroll(selenium, website):
    selenium.get(urljoin(website, "/browsers_scroll.htm"))
    img_browsers = selenium.find_element(By.ID, "browsers")
    selenium.execute_script("arguments[0].scrollIntoView();", img_browsers)
    selenium.find_image("tests/html/browser_edge.png").click()
    assert (
        selenium.find_element(By.ID, "message").text == "edge1"
    ), "click on element failed"



from selenium import webdriver
from selenium.webdriver.remote.webelement import WebElement

from typing import Union


def move_to(self: WebElement):
    """
    Move the mouse to the center of the specified element.
    """
    __web_element_move_and_click(self, None, None, False)


def move_at(self: WebElement, xoffset: int, yoffset: int):
    """
    Move the mouse by an offset of the specified element.
        Offsets are relative to the top-left corner of the element.

    :Args:
        - xoffset: X offset to move to.
        - yoffset: Y offset to move to.
    """
    __web_element_move_and_click(self, xoffset, yoffset, False)


def click_at(self: WebElement, xoffset: int, yoffset: int):
    """
    Move the mouse by an offset of the specified element and click.
        Offsets are relative to the top-left corner of the element.

    :Args:
        - xoffset: X offset to move to.
        - yoffset: Y offset to move to.
    """
    __web_element_move_and_click(self, xoffset, yoffset, True)


def __web_element_move_and_click(
    webelement: WebElement,
    xoffset: Union[int, None],
    yoffset: Union[int, None],
    click: bool,
):
    action = webdriver.common.action_chains.ActionChains(webelement.parent)
    if xoffset is None or yoffset is None:
        action.move_to_element(webelement)
    else:
        action.move_to_element_with_offset(webelement, xoffset, yoffset)
    if click:
        action.click()
    action.perform()



import selenium
from niobium.patch import (
    patch_image,
    patch_timeout,
    patch_webelement_action,
    patch_webelement_wait,
)

patch_image()
patch_timeout()
patch_webelement_action()
patch_webelement_wait()

__all__ = ["selenium"]

__version__ = "1.0.0"



from urllib.parse import urljoin

import pytest
from selenium.common.exceptions import (
    NoSuchElementException,
    ElementNotVisibleException,
    ElementNotInteractableException,
)
from selenium.webdriver.common.by import By


def test_web_element_need_wait(selenium, website):
    selenium.get(urljoin(website, "/browsers_delay.htm"))
    with pytest.raises(ElementNotVisibleException):
        selenium.find_element(By.ID, "button_displayed_enabled").wait()
    with pytest.raises(ElementNotInteractableException):
        selenium.find_element(By.ID, "button_displayed_disabled").wait(
            displayed=None, enabled=True
        )
    with pytest.raises(NoSuchElementException):
        selenium.find_element(By.ID, "button_always_displayed_enabled").wait(
            displayed=False, enabled=None
        )
    with pytest.raises(NoSuchElementException):
        selenium.find_element(By.ID, "button_always_displayed_enabled").wait(
            displayed=None, enabled=False
        )


def test_web_element_wait_displayed_implict(selenium, website):
    selenium.implicitly_wait(6)
    selenium.get(urljoin(website, "/browsers_delay.htm"))
    selenium.find_element(By.ID, "button_displayed_enabled").wait(
        displayed=True, enabled=None
    )


def test_web_element_wait_enabled_implict(selenium, website):
    selenium.implicitly_wait(6)
    selenium.get(urljoin(website, "/browsers_delay.htm"))
    selenium.find_element(By.ID, "button_displayed_enabled").wait(
        displayed=None, enabled=True
    )


def test_web_element_wait_displayed_explicit(selenium, website):
    selenium.get(urljoin(website, "/browsers_delay.htm"))
    selenium.find_element(By.ID, "button_displayed_enabled").wait(
        displayed=True, enabled=None, timeout=6
    )


def test_web_element_wait_enabled_explicit(selenium, website):
    selenium.implicitly_wait(6)
    selenium.get(urljoin(website, "/browsers_delay.htm"))
    selenium.find_element(By.ID, "button_displayed_enabled").wait(None, True, 6)


def test_web_element_wait_displayed_then_click(selenium, website):
    selenium.implicitly_wait(6)
    selenium.get(urljoin(website, "/browsers_delay.htm"))
    selenium.find_element(By.ID, "button_displayed_enabled").wait().click()
    assert (
        selenium.find_element(By.ID, "message").text == "button_displayed_enabled"
    ), "click on element failed"



import os

import cv2

from niobium.find_image import match_template


img_dir = os.path.dirname(os.path.realpath(__file__))


def test_find_match_template_zero():
    img_src = cv2.imread(os.path.join(img_dir, "html/browsers.png"))
    img_template = cv2.imread(os.path.join(img_dir, "html/browser0.png"))
    res = match_template(img_src, img_template)
    assert len(res) == 0


def test_find_match_template_one():
    img_src = cv2.imread(os.path.join(img_dir, "html/browsers.png"))
    img_template = cv2.imread(os.path.join(img_dir, "html/browser_edge.png"))
    res = match_template(img_src, img_template)
    assert len(res) == 1
    assert (150, 27) == res[0]


def test_find_match_template_three():
    img_src = cv2.imread(os.path.join(img_dir, "html/browsers.png"))
    img_template = cv2.imread(os.path.join(img_dir, "html/browser_firefox.png"))
    res = set(match_template(img_src, img_template))
    assert len(res) == 3
    assert (321, 21) in res
    assert (478, 153) in res
    assert (27, 157) in res


def test_find_match_template_threshold():
    img_src = cv2.imread(os.path.join(img_dir, "html/browsers.png"))
    img_template = cv2.imread(os.path.join(img_dir, "html/browser_edge_degraded.png"))
    res = match_template(img_src, img_template)
    assert len(res) == 0
    res = match_template(img_src, img_template, 0.8)
    assert len(res) == 1
    assert (150, 27) == res[0]



import io

import cv2
import numpy as np
from PIL import Image
from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webdriver import WebDriver

from niobium.image_element import ImageElement
from niobium.timeout import ImplicitWait


def match_template(img_src, img_template, threshold=0.9):
    # we remove alpha channel if exists
    img_src = img_src[:, :, :3]
    img_template = img_template[:, :, :3]

    res = cv2.matchTemplate(img_src, img_template, cv2.TM_CCOEFF_NORMED)

    _, max_val, _, _ = cv2.minMaxLoc(res)

    if max_val >= threshold:
        # change threshold in order to return only the different instances
        threshold = max(threshold, max_val - 0.01)

        loc = np.where(res >= threshold)

        return list(zip(*loc[::-1]))
    else:
        return []


def find_images(self: WebDriver, filename: str) -> list[ImageElement]:
    """
    Locate all the occurence of an image in the webpage.

    :Args:
        - filename: The path to the image to search (image shall be in PNG format).

    :Returns:
        A list of ImageElement.
    """
    template = cv2.imread(filename, cv2.IMREAD_UNCHANGED)
    if template is None:
        raise WebDriverException(f"Unable to read {filename}")
    template_height, template_width, _ = template.shape

    webpage_png = self.get_screenshot_as_png()
    webpage_img = Image.open(io.BytesIO(webpage_png))
    webpage: np.ndarray = np.asarray(webpage_img, dtype=np.float32).astype(np.uint8)
    webpage = cv2.cvtColor(webpage, cv2.COLOR_BGR2RGB)

    return [
        ImageElement(self, loc[0], loc[1], template_width, template_height)
        for loc in match_template(webpage, template)
    ]


def find_image(self: WebDriver, filename: str) -> ImageElement:
    """
    Locate an image in the webpage.

    :Args:
        - filename: The path to the image to search (image shall be in PNG format).

    :Returns:
        An ImageElement.

    :Raises:
        - NoSuchElementException - if the element wasn't found
    """
    implicit_wait = ImplicitWait()
    implicit_wait.start()

    elements = find_images(self, filename)

    while len(elements) == 0 and not implicit_wait.max_time_exceeded:
        elements = find_images(self, filename)

    if len(elements) == 0:
        raise NoSuchElementException(
            'Unable to locate element: [image="{filename}"]'.format(filename=filename)
        )
    else:
        return elements[0]



from urllib.parse import urljoin

from selenium.webdriver.common.by import By


def test_find_images_no_element(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    assert len(selenium.find_images("tests/html/browser0.png")) == 0


def test_find_images_one_element(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    assert len(selenium.find_images("tests/html/browser_edge.png")) == 1


def test_find_images_three_elements(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    assert len(selenium.find_images("tests/html/browser_firefox.png")) == 3


def test_find_images_one_element_click(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    elements = selenium.find_images("tests/html/browser_edge.png")
    assert len(elements) == 1
    elements[0].click()
    assert (
        selenium.find_element(By.ID, "message").text == "edge1"
    ), "click on element failed"


def test_find_images_three_elements_click(selenium, website):
    selenium.get(urljoin(website, "/browsers.htm"))
    elements = selenium.find_images("tests/html/browser_firefox.png")
    assert len(elements) == 3
    messages = []
    for element in elements:
        element.click()
        messages.append(selenium.find_element(By.ID, "message").text)
    messages.sort()
    assert messages == ["firefox1", "firefox2", "firefox3"], "click on element failed"



import time

from selenium import webdriver
from selenium.webdriver.remote.webdriver import WebDriver


class ImplicitWait(object):
    time_to_wait: float = 0

    def __init__(self):
        self.start_time = None

    def start(self):
        self.start_time = time.time()

    @property
    def max_time_exceeded(self):
        return (time.time() - self.start_time) > self.time_to_wait


_real_implicitly_wait = webdriver.remote.webdriver.WebDriver.implicitly_wait


def _patch_implicitly_wait(self: WebDriver, time_to_wait: float):
    ImplicitWait.time_to_wait = time_to_wait
    return _real_implicitly_wait(self, time_to_wait)


class ExplicitWait(object):
    def __init__(self, time_to_wait):
        self.start_time = None
        self.time_to_wait = time_to_wait

    def start(self):
        self.start_time = time.time()

    @property
    def max_time_exceeded(self):
        return (time.time() - self.start_time) > self.time_to_wait



import time

from niobium.timeout import ImplicitWait


def test_implicit_wait_max_time_exceeded():
    ImplicitWait.time_to_wait = 1
    implicit_wait = ImplicitWait()
    implicit_wait.start()
    assert not implicit_wait.max_time_exceeded
    time.sleep(1)
    assert implicit_wait.max_time_exceeded



import time

from selenium.common.exceptions import (
    ElementNotVisibleException,
    ElementNotInteractableException,
    NoSuchElementException,
)
from selenium.webdriver.remote.webelement import WebElement

from niobium.timeout import ImplicitWait, ExplicitWait


def wait(
    self: WebElement, displayed: bool = True, enabled: bool = True, timeout: bool = None
):
    """
    Wait until expected conditions.

    :Args:
        - displayed (bool - default: True): Wait until the element is displayed.
        - enabled  (bool - default: True): Wait until the element is enabled.
        - timeout  (bool - default: None): Timeout in seconds, implicit timeout if None.

    :Returns:
        The element itself.

    :Raises:
         - ElementNotVisibleException - if element is not visible.
         - ElementNotInteractableException - if element is not enabled.
         - NoSuchElementException - if expected conditions not satisfied.
    """
    wait_timer = ExplicitWait(timeout) if timeout is not None else ImplicitWait()
    wait_timer.start()

    once = True
    while (not wait_timer.max_time_exceeded) or once:
        once = False
        if ((displayed is None) or (displayed == self.is_displayed())) and (
            (enabled is None) or (enabled == self.is_enabled())
        ):
            return self
        time.sleep(0.2)

    if displayed:
        raise ElementNotVisibleException()
    elif enabled:
        raise ElementNotInteractableException()
    else:
        raise NoSuchElementException()



from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from typing import Dict
from typing import Union


class ImageElement(object):
    def __init__(
        self, parent: WebDriver, x: int, y: int, width: int, height: int
    ) -> None:
        """
        Create a new ImageElement.

        :Args:
         - parent: The WebDriver.
         - x: location of the element on the X axis.
         - y: location of the element on the Y axis.
         - width: the width of the element.
         - height: the height of the element.
        """
        self._parent = parent
        self._location = {"x": x, "y": y}
        self._size = {"width": width, "height": height}
        self._rect = {"x": x, "y": y, "width": width, "height": height}

    @property
    def parent(self) -> WebDriver:
        """Internal reference to the WebDriver instance this element was found from."""
        return self._parent

    @property
    def location(self) -> Dict[str, int]:
        """The location of the element in the renderable canvas."""
        return self._location

    @property
    def size(self) -> Dict[str, int]:
        """The size of the element."""
        return self._size

    @property
    def rect(self) -> Dict[str, int]:
        """A dictionary with the size and location of the element."""
        return self._rect

    def move_to(self) -> None:
        """
        Move the mouse to the center of the specified element.
        """
        self.__move_and_click(None, None, False)

    def move_at(self, xoffset: int, yoffset: int) -> None:
        """
        Move the mouse by an offset of the specified element.
           Offsets are relative to the top-left corner of the element.

        :Args:
         - xoffset: X offset to move to.
         - yoffset: Y offset to move to.
        """
        self.__move_and_click(xoffset, yoffset, False)

    def click(self) -> None:
        """
        Move the mouse to the center of the specified element and click.
        """
        self.__move_and_click(None, None, True)

    def click_at(self, xoffset: int, yoffset: int) -> None:
        """
        Move the mouse by an offset of the specified element and click.
           Offsets are relative to the top-left corner of the element.

        :Args:
         - xoffset: X offset to move to.
         - yoffset: Y offset to move to.
        """
        self.__move_and_click(xoffset, yoffset, True)

    def __move_and_click(
        self,
        xoffset: Union[int, None] = None,
        yoffset: Union[int, None] = None,
        click: bool = False,
    ) -> None:
        if xoffset is None:
            xoffset = self.location["x"] + int(self.size["width"] / 2)
        else:
            xoffset = self.location["x"] + xoffset

        if yoffset is None:
            yoffset = self.location["y"] + int(self.size["height"] / 2)
        else:
            yoffset = self.location["y"] + yoffset

        action = webdriver.common.action_chains.ActionChains(self.parent)
        offset_element = self.__get_element_top_left()
        action.move_to_element_with_offset(offset_element, xoffset, yoffset)
        if click:
            action.click()
        action.perform()

    def __get_element_top_left(self):
        # create an empty element at the corner top left of the webpage in order to be
        # able to click in the good position even if the page is scrolled
        javacript_create_topleft_element = """
        if(! document.getElementById("niobium-topleft-elt"))
        {
            var topleft_elt = document.createElement("div");
            topleft_elt.style.top=0;
            topleft_elt.style.left=0;
            topleft_elt.style.height=0;
            topleft_elt.style.width=0;
            topleft_elt.style.margin=0;
            topleft_elt.style.padding=0;
            topleft_elt.style.position="fixed";
            topleft_elt.id="niobium-topleft-elt";
            document.body.appendChild(topleft_elt);
        }
        """
        self.parent.execute_script(javacript_create_topleft_element)
        return self.parent.find_element(By.ID, "niobium-topleft-elt")



from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement

from niobium.find_image import find_image
from niobium.find_image import find_images
from niobium.timeout import _patch_implicitly_wait
from niobium.webelement_action import click_at, move_to, move_at
from niobium.webelement_wait import wait


def patch_image():

    if not hasattr(WebDriver, "find_image"):
        WebDriver.find_image = find_image

    if not hasattr(WebDriver, "find_images"):
        WebDriver.find_images = find_images


def patch_timeout():
    WebDriver.implicitly_wait = _patch_implicitly_wait


def patch_webelement_action():

    if not hasattr(WebElement, "click_at"):
        WebElement.click_at = click_at

    if not hasattr(WebElement, "move_to"):
        WebElement.move_to = move_to

    if not hasattr(WebElement, "move_at"):
        WebElement.move_at = move_at


def patch_webelement_wait():
    if not hasattr(WebElement, "wait"):
        WebElement.wait = wait
